"""Define BaseModuleDispatcher class."""

from __future__ import annotations

import abc

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from collections.abc import Sequence

from pytest_ansible.errors import AnsibleModuleError


class BaseModuleDispatcher(abc.ABC):
    """The module dispatcher is responsible for running ansible modules.

    Attributes:
        required_kwargs: The required keyword arguments for the dispatcher.
    """

    required_kwargs: Sequence[str] = ("inventory",)

    def __init__(self, **kwargs) -> None:  # type: ignore[no-untyped-def]  # noqa: ANN003
        """Save provided keyword arguments and assert required values have been provided."""
        self.options = kwargs

        # Assert the expected kwargs were provided
        self.check_required_kwargs(**kwargs)  # type: ignore[no-untyped-call]

    def __len__(self) -> int:
        """Return the number of hosts that match the `host_pattern`."""
        try:
            extra_inventory_hosts = self.options["extra_inventory_manager"].list_hosts(
                self.options["host_pattern"],
            )
        except KeyError:
            extra_inventory_hosts = []
        return len(
            self.options["inventory_manager"].list_hosts(self.options["host_pattern"]),
        ) + len(extra_inventory_hosts)

    def __contains__(self, item) -> bool:  # type: ignore[no-untyped-def]  # noqa: ANN001
        """Return the whether the inventory or extra_inventory contains a host matching the provided `item`."""  # noqa: E501
        try:
            extra_inventory_hosts = self.options["extra_inventory_manager"].list_hosts(
                item,
            )
        except KeyError:
            extra_inventory_hosts = []
        return (
            len(self.options["inventory_manager"].list_hosts(item)) + len(extra_inventory_hosts)
        ) > 0

    def __getattr__(self, name):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN204
        """Run the ansible module matching the provided `name`.

        Raise `AnsibleModuleError` when no such module exists.
        """
        resolved = self.has_module(name)
        if not resolved:
            msg = f"The module {name} was not found in configured module paths."
            raise AnsibleModuleError(msg)
        self.options["module_name"] = resolved
        return self._run

    def check_required_kwargs(self, **kwargs):  # type: ignore[no-untyped-def]  # noqa: ANN003, ANN201, ARG002
        """Raise a TypeError if any required kwargs are missing."""
        for kwarg in self.required_kwargs:
            if kwarg not in self.options:
                msg = f"Missing required keyword argument '{kwarg}'"
                raise TypeError(msg)

    @abc.abstractmethod
    def has_module(self: BaseModuleDispatcher, name: str) -> str:
        """Return whether ansible provides the requested module.

        Must be implemented by a sub-class.
        """
        return ""

    @abc.abstractmethod
    def _run(self, *args, **kwargs):  # type: ignore[no-untyped-def]  # noqa: ANN002, ANN003, ANN202
        """Raise a runtime error, unless implemented by sub-classes."""
        msg = "Must be implemented by a sub-class"
        raise RuntimeError(msg)



"""Test the host manager."""

from __future__ import annotations

import pytest

from .conftest import (
    ALL_EXTRA_HOSTS,
    ALL_HOSTS,
    EXTRA_HOST_POSITIVE_PATTERNS,
    NEGATIVE_HOST_PATTERNS,
    NEGATIVE_HOST_SLICES,
    POSITIVE_HOST_PATTERNS,
    POSITIVE_HOST_SLICES,
)


pytestmark = [
    pytest.mark.unit,
]


@pytest.mark.parametrize(
    "include_extra_inventory",
    (True, False),
)
def test_host_manager_len(hosts, include_extra_inventory):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201, D103
    _hosts = hosts(include_extra_inventory=include_extra_inventory)
    assert len(_hosts) == len(ALL_HOSTS) + len(
        ALL_EXTRA_HOSTS if include_extra_inventory else [],
    )


@pytest.mark.parametrize(
    "include_extra_inventory",
    (True, False),
)
def test_host_manager_keys(hosts, include_extra_inventory):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201, D103
    _hosts = hosts(include_extra_inventory=include_extra_inventory)
    sorted_keys = _hosts.keys()
    sorted_keys.sort()
    for key in sorted_keys:
        assert key in ALL_HOSTS + (ALL_EXTRA_HOSTS if include_extra_inventory else [])


@pytest.mark.parametrize(
    ("host_pattern", "num_hosts"),
    POSITIVE_HOST_PATTERNS + EXTRA_HOST_POSITIVE_PATTERNS,
)
@pytest.mark.parametrize(
    "include_extra_inventory",
    (True, False),
)
def test_host_manager_contains(host_pattern, num_hosts, hosts, include_extra_inventory):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201, ARG001, D103
    _hosts = hosts(include_extra_inventory=include_extra_inventory)
    if not include_extra_inventory and host_pattern.startswith("extra"):
        assert host_pattern not in _hosts, f"{host_pattern} in hosts"
    else:
        assert host_pattern in _hosts, f"{host_pattern} not in hosts"


@pytest.mark.parametrize(
    ("host_pattern", "num_hosts"),
    NEGATIVE_HOST_PATTERNS,
)
@pytest.mark.parametrize("include_extra_inventory", (True, False))
def test_host_manager_not_contains(  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    host_pattern,  # noqa: ANN001
    num_hosts,  # noqa: ANN001, ARG001
    hosts,  # noqa: ANN001
    include_extra_inventory,  # noqa: ANN001
):
    _hosts = hosts(include_extra_inventory=include_extra_inventory)
    assert host_pattern not in _hosts


@pytest.mark.parametrize(
    ("host_pattern", "num_hosts"),
    POSITIVE_HOST_PATTERNS + EXTRA_HOST_POSITIVE_PATTERNS,
)
@pytest.mark.parametrize(
    "include_extra_inventory",
    (True, False),
)
def test_host_manager_getitem(host_pattern, num_hosts, hosts, include_extra_inventory):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201, ARG001, D103
    _hosts = hosts(include_extra_inventory=include_extra_inventory)
    if not include_extra_inventory and host_pattern.startswith("extra"):
        assert host_pattern not in _hosts
    else:
        assert _hosts[host_pattern]


@pytest.mark.parametrize(
    ("host_pattern", "num_hosts"),
    NEGATIVE_HOST_PATTERNS,
)
@pytest.mark.parametrize(
    "include_extra_inventory",
    (True, False),
)
def test_host_manager_not_getitem(  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    host_pattern,  # noqa: ANN001
    num_hosts,  # noqa: ANN001, ARG001
    hosts,  # noqa: ANN001
    include_extra_inventory,  # noqa: ANN001
):
    _hosts = hosts(include_extra_inventory=include_extra_inventory)
    with pytest.raises(KeyError):
        assert _hosts[host_pattern]


@pytest.mark.parametrize(
    ("host_pattern", "num_hosts"),
    POSITIVE_HOST_PATTERNS + EXTRA_HOST_POSITIVE_PATTERNS,
)
@pytest.mark.parametrize(
    "include_extra_inventory",
    (True, False),
)
def test_host_manager_getattr(host_pattern, num_hosts, hosts, include_extra_inventory):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201, ARG001, D103
    _hosts = hosts(include_extra_inventory=include_extra_inventory)
    if not include_extra_inventory and host_pattern.startswith("extra"):
        assert not hasattr(_hosts, host_pattern)
    else:
        assert hasattr(_hosts, host_pattern)


@pytest.mark.parametrize(
    ("host_slice", "num_hosts"),
    POSITIVE_HOST_SLICES,
)
@pytest.mark.parametrize(
    "include_extra_inventory",
    (True, False),
)
def test_host_manager_slice(host_slice, num_hosts, hosts, include_extra_inventory):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201, D103
    _hosts = hosts(include_extra_inventory=include_extra_inventory)
    assert (
        len(_hosts[host_slice]) == num_hosts[include_extra_inventory]
    ), f"{len(_hosts[host_slice])} != {num_hosts} for {host_slice}"


# pylint: disable=pointless-statement
@pytest.mark.parametrize(
    "host_slice",
    NEGATIVE_HOST_SLICES,
)
@pytest.mark.parametrize(
    "include_extra_inventory",
    (True, False),
)
def test_host_manager_not_slice(host_slice, hosts, include_extra_inventory):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201, D103
    _hosts = hosts(include_extra_inventory=include_extra_inventory)
    with pytest.raises(KeyError):
        _hosts[host_slice]


@pytest.mark.parametrize(
    ("host_pattern", "num_hosts"),
    NEGATIVE_HOST_PATTERNS,
)
@pytest.mark.parametrize(
    "include_extra_inventory",
    (True, False),
)
def test_host_manager_not_getattr(  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    host_pattern,  # noqa: ANN001
    num_hosts,  # noqa: ANN001, ARG001
    hosts,  # noqa: ANN001
    include_extra_inventory,  # noqa: ANN001
):
    _hosts = hosts(include_extra_inventory=include_extra_inventory)
    assert not hasattr(_hosts, host_pattern)
    with pytest.raises(AttributeError):
        getattr(_hosts, host_pattern)


def test_defaults(request):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201, D103
    from ansible.constants import DEFAULT_TRANSPORT  # pylint: disable=no-name-in-module

    plugin = request.config.pluginmanager.getplugin("ansible")
    hosts = plugin.initialize(config=request.config, request=request)

    assert "connection" in hosts.options
    assert hosts.options["connection"] == DEFAULT_TRANSPORT



"""PyTest Ansible Plugin."""

from __future__ import annotations

import contextlib
import logging
import subprocess

from typing import TYPE_CHECKING

import ansible
import ansible.constants
import ansible.errors
import ansible.utils
import ansible.utils.display
import pytest

from pytest_ansible.fixtures import (
    ansible_facts,
    fixture_ansible_adhoc,
    fixture_ansible_module,
    localhost,
)
from pytest_ansible.host_manager.utils import get_host_manager

from .molecule import HAS_MOLECULE, MoleculeFile, MoleculeScenario
from .units import inject, inject_only


if TYPE_CHECKING:
    from pathlib import Path

    from _pytest.nodes import Node

logger = logging.getLogger(__name__)

# Silence linters for imported fixtures
# pylint: disable=pointless-statement, no-member
(fixture_ansible_adhoc, fixture_ansible_module, ansible_facts, localhost)  # noqa: B018

log_map = {
    0: logging.CRITICAL,
    1: logging.ERROR,
    2: logging.WARNING,
    3: logging.INFO,
    4: logging.DEBUG,
}
OUR_FIXTURES = ("ansible_adhoc", "ansible_module", "ansible_facts")


def pytest_addoption(parser):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201
    """Add options to control ansible."""
    group = parser.getgroup("pytest-ansible")
    group.addoption(
        "--inventory",
        "--ansible-inventory",
        action="store",
        dest="ansible_inventory",
        default=ansible.constants.DEFAULT_HOST_LIST,
        metavar="ANSIBLE_INVENTORY",
        help="ansible inventory file URI (default: %(default)s)",
    )
    group.addoption(
        "--extra-inventory",
        "--ansible-extra-inventory",
        action="store",
        dest="ansible_extra_inventory",
        default=None,
        metavar="ANSIBLE_EXTRA_INVENTORY",
        help="ansible extra inventory file URI (default: %(default)s)",
    )
    group.addoption(
        "--host-pattern",
        "--ansible-host-pattern",
        action="store",
        dest="ansible_host_pattern",
        default=None,
        metavar="ANSIBLE_HOST_PATTERN",
        help="ansible host pattern (default: %(default)s)",
    )
    group.addoption(
        "--limit",
        "--ansible-limit",
        action="store",
        dest="ansible_subset",
        default=ansible.constants.DEFAULT_SUBSET,
        metavar="ANSIBLE_SUBSET",
        help="further limit selected hosts to an additional pattern",
    )
    group.addoption(
        "--connection",
        "--ansible-connection",
        action="store",
        dest="ansible_connection",
        default=ansible.constants.DEFAULT_TRANSPORT,
        help="connection type to use (default: %(default)s)",
    )
    group.addoption(
        "--user",
        "--ansible-user",
        action="store",
        dest="ansible_user",
        default=ansible.constants.DEFAULT_REMOTE_USER,
        help="connect as this user (default: %(default)s)",
    )
    group.addoption(
        "--check",
        "--ansible-check",
        action="store_true",
        dest="ansible_check",
        default=False,
        help="don't make any changes; instead, try to predict some of the changes that may occur",
    )
    group.addoption(
        "--module-path",
        "--ansible-module-path",
        action="store",
        dest="ansible_module_path",
        default=ansible.constants.DEFAULT_MODULE_PATH,
        help="specify path(s) to module library (default: %(default)s)",
    )

    # become privilege escalation
    group.addoption(
        "--become",
        "--ansible-become",
        action="store_true",
        dest="ansible_become",
        default=ansible.constants.DEFAULT_BECOME,
        help="run operations with become, nopasswd implied (default: %(default)s)",
    )
    group.addoption(
        "--become-method",
        "--ansible-become-method",
        action="store",
        dest="ansible_become_method",
        default=ansible.constants.DEFAULT_BECOME_METHOD,
        help="privilege escalation method to use (default: %(default)s)",
    )

    group.addoption(
        "--become-user",
        "--ansible-become-user",
        action="store",
        dest="ansible_become_user",
        default=ansible.constants.DEFAULT_BECOME_USER,
        help="run operations as this user (default: %(default)s)",
    )
    group.addoption(
        "--ask-become-pass",
        "--ansible-ask-become-pass",
        action="store",
        dest="ansible_ask_become_pass",
        default=ansible.constants.DEFAULT_BECOME_ASK_PASS,
        help="ask for privilege escalation password (default: %(default)s)",
    )
    group.addoption(
        "--ansible-unit-inject-only",
        action="store_true",
        default=False,
        help="Enable support for ansible collection unit tests by only injecting existing ANSIBLE_COLLECTIONS_PATH.",  # noqa: E501
    )
    group.addoption(
        "--molecule",
        action="store_true",
        default=False,
        help="Enable support for running discovered molecule scenarios from pytest.",
    )
    group.addoption(
        "--molecule_unavailable_driver",
        action="store",
        default=None,
        help="What marker to add to molecule scenarios when driver is ",
    )
    group.addoption(
        "--molecule_base_config",
        action="store",
        default=None,
        help="Path to the molecule base config file. The value of this option is ",
    )
    group.addoption(
        "--skip_no_git_change",
        action="store",
        default=None,
        help="Commit to use as a reference for this test. If the role wasn't",
    )
    # Add github marker to --help
    parser.addini("ansible", "Ansible integration", "args")


def pytest_configure(config):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201
    """Validate --ansible-* parameters."""
    config.addinivalue_line("markers", "ansible(**kwargs): Ansible integration")

    # Enable connection debugging
    if config.option.verbose > 0:
        if hasattr(ansible.utils, "VERBOSITY"):
            ansible.utils.VERBOSITY = int(config.option.verbose)
        else:
            ansible.utils.display.verbosity = int(config.option.verbose)

    # Configure the logger.
    level = log_map.get(config.option.verbose)
    logging.basicConfig(level=level)
    logging.debug("Logging initialized")

    assert config.pluginmanager.register(PyTestAnsiblePlugin(config), "ansible")  # noqa: S101

    if config.option.ansible_unit_inject_only:
        inject_only()
    else:
        start_path = config.invocation_params.dir
        inject(start_path)


def pytest_collect_file(
    file_path: Path | None,
    parent: pytest.Collector,
) -> Node | None:
    """Transform each found molecule.yml into a pytest test."""
    if not parent.config.option.molecule:
        return None
    if not HAS_MOLECULE:
        pytest.exit(
            f"molecule not installed or found, unable to collect test {file_path}",
        )
        return None
    if file_path and file_path.is_symlink():
        return None
    if file_path and file_path.name == "molecule.yml":
        return MoleculeFile.from_parent(path=file_path, parent=parent)
    return None


def pytest_generate_tests(metafunc):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201
    """Generate tests when specific `ansible_*` fixtures are used by tests."""
    if "ansible_host" in metafunc.fixturenames:
        # assert required --ansible-* parameters were used
        PyTestAnsiblePlugin.assert_required_ansible_parameters(metafunc.config)  # type: ignore[no-untyped-call]
        try:
            plugin = metafunc.config.pluginmanager.getplugin("ansible")
            hosts = plugin.initialize(
                config=plugin.config,
                pattern=metafunc.config.getoption("ansible_host_pattern"),
            )
        except ansible.errors.AnsibleError as exception:
            raise pytest.UsageError(exception)  # noqa: B904

        # Return the host name as a string
        metafunc.parametrize("ansible_host", iter(hosts[h] for h in hosts))

    if "ansible_group" in metafunc.fixturenames:
        # assert required --ansible-* parameters were used
        PyTestAnsiblePlugin.assert_required_ansible_parameters(metafunc.config)  # type: ignore[no-untyped-call]
        try:
            plugin = metafunc.config.pluginmanager.getplugin("ansible")
            hosts = plugin.initialize(
                config=plugin.config,
                pattern=metafunc.config.getoption("ansible_host_pattern"),
            )
        except ansible.errors.AnsibleError as exception:
            raise pytest.UsageError(exception)  # noqa: B904
        groups = hosts.options["inventory_manager"].list_groups()
        extra_groups = hosts.get_extra_inventory_groups()
        # Return the group name as a string
        metafunc.parametrize("ansible_group", iter(hosts[g] for g in groups))
        metafunc.parametrize("ansible_group", iter(hosts[g] for g in extra_groups))

    if "molecule_scenario" in metafunc.fixturenames:
        if not HAS_MOLECULE:
            pytest.exit("molecule not installed or found.")

        # Find all molecule scenarios not gitignored
        rootpath = metafunc.config.rootpath

        scenarios = []

        candidates = list(rootpath.glob("**/molecule/*/molecule.yml"))
        command = ["git", "check-ignore", *candidates]
        with contextlib.suppress(subprocess.CalledProcessError, FileNotFoundError):
            proc = subprocess.run(
                args=command,
                capture_output=True,
                check=True,
                text=True,
                shell=False,
            )

        try:
            ignored = proc.stdout.splitlines()
            scenario_paths = [
                candidate for candidate in candidates if str(candidate) not in ignored
            ]
        except NameError:
            scenario_paths = candidates

        for fs_entry in scenario_paths:
            scenario = fs_entry.parent
            molecule_parent = scenario.parent.parent
            scenarios.append(
                MoleculeScenario(
                    parent_directory=molecule_parent,
                    name=scenario.name,
                    test_id=f"{molecule_parent.name}-{scenario.name}",
                ),
            )
        if not scenarios:
            pytest.exit(f"No molecule scenarios found in: {rootpath}")
        metafunc.parametrize(
            "molecule_scenario",
            scenarios,
            ids=[scenario.test_id for scenario in scenarios],
        )


class PyTestAnsiblePlugin:
    """Ansible PyTest Plugin Class."""

    def __init__(self, config) -> None:  # type: ignore[no-untyped-def]  # noqa: ANN001
        """Initialize plugin."""
        self.config = config

    def pytest_report_header(self):  # type: ignore[no-untyped-def]  # noqa: ANN201
        """Return the version of ansible."""
        return f"ansible: {ansible.__version__}"

    def pytest_collection_modifyitems(self, session, config, items):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201, ARG002
        """Validate --ansible-* parameters."""
        uses_ansible_fixtures = False
        for item in items:
            if not hasattr(item, "fixturenames"):
                continue

            for fixture_name in item.fixturenames:
                if fixture_name in OUR_FIXTURES:
                    uses_ansible_fixtures = True
                    break

                # ignore any normal fixtures that have definitions to avoid miss activations
                if (
                    hasattr(item, "_fixtureinfo")
                    and hasattr(item._fixtureinfo, "name2fixturedefs")  # noqa: SLF001
                    and fixture_name in item._fixtureinfo.name2fixturedefs  # noqa: SLF001
                ):
                    continue
                if fixture_name == "request":
                    continue  # reserved name, from pytest
                logger.error(
                    "Found %s fixture which seem to have no definition.",
                    fixture_name,
                )

        if uses_ansible_fixtures:
            # assert required --ansible-* parameters were used
            self.assert_required_ansible_parameters(config)  # type: ignore[no-untyped-call]

    def _load_ansible_config(self, config):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN202
        """Load ansible configuration from command-line."""
        option_names = [
            "ansible_inventory",
            "ansible_extra_inventory",
            "ansible_host_pattern",
            "ansible_connection",
            "ansible_user",
            "ansible_module_path",
            "ansible_become",
            "ansible_become_method",
            "ansible_become_user",
            "ansible_ask_become_pass",
            "ansible_subset",
        ]

        kwargs = {}

        # Load command-line supplied values
        for key in option_names:
            short_key = key[8:]
            kwargs[short_key] = config.getoption(key)

        # normalize ansible.ansible_become options
        kwargs["become"] = kwargs.get("become") or ansible.constants.DEFAULT_BECOME
        kwargs["become_user"] = kwargs.get("become_user") or ansible.constants.DEFAULT_BECOME_USER
        kwargs["ask_become_pass"] = (
            kwargs.get("ask_become_pass") or ansible.constants.DEFAULT_BECOME_ASK_PASS
        )

        return kwargs

    def _load_request_config(self, request):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN202
        """Load ansible configuration from decorator kwargs."""
        kwargs = {}

        # Override options from @pytest.mark.ansible
        marker = request.node.get_closest_marker("ansible")
        if marker:
            kwargs = marker.kwargs

        return kwargs

    def initialize(self, config=None, request=None, **kwargs):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN003, ANN201
        """Return an initialized Ansible Host Manager instance."""
        ansible_cfg = {}
        # merge command-line configuration options
        if config is not None:
            ansible_cfg.update(self._load_ansible_config(config))  # type: ignore[no-untyped-call]
        # merge pytest request configuration options
        if request is not None:
            ansible_cfg.update(self._load_request_config(request))  # type: ignore[no-untyped-call]
        # merge in provided kwargs
        ansible_cfg.update(kwargs)
        return get_host_manager(**ansible_cfg)

    @staticmethod
    def assert_required_ansible_parameters(config):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN205
        """Assert whether the required --ansible-* parameters were provided."""
        errors = []

        # Verify --ansible-host-pattern was provided
        ansible_hostname = config.getoption("ansible_host_pattern")
        if ansible_hostname is None or ansible_hostname == "":
            errors.append(
                "Missing required parameter --ansible-host-pattern/--host-pattern",
            )

        # NOTE: I don't think this will ever catch issues since ansible_inventory
        # defaults to '/etc/ansible/hosts'
        # Verify --ansible-inventory was provided
        ansible_inventory = config.getoption("ansible_inventory")
        if ansible_inventory is None or ansible_inventory == "":
            errors.append(
                "Unable to find an inventory file, specify one with the --ansible-inventory/--inventory "  # noqa: E501
                "parameter.",
            )

        if errors:
            raise pytest.UsageError(*errors)



"""Global fixtures for pytest tests."""

from __future__ import annotations

import pytest

from pytest_ansible.host_manager.utils import get_host_manager


try:
    from ansible.utils import context_objects as co
except ImportError:
    # if it does not exist because of old version of ansible, we don't need it
    co = None

pytest_plugins = ["pytester"]

ALL_HOSTS = ["another_host", "localhost", "yet_another_host"]

ALL_EXTRA_HOSTS = ["extra-hosts", "extra-hosts_extra", "yet_another_extra-hosts"]

POSITIVE_HOST_PATTERNS = [
    # True means we are including the extra inventory, False not
    pytest.param("all", {True: 6, False: 3}, id="0"),
    pytest.param("*", {True: 6, False: 3}, id="1"),
    pytest.param("localhost", {True: 2, False: 1}, id="2"),
    pytest.param("local*", {True: 1, False: 1}, id="3"),
    pytest.param("local*,&*host", {True: 1, False: 1}, id="4"),
    pytest.param("!localhost", {True: 5, False: 2}, id="5"),
    pytest.param("all[0]", {True: 2, False: 1}, id="6"),
    pytest.param("all[-1]", {True: 2, False: 1}, id="7"),
    pytest.param(
        "*[0:1]",
        {True: 4, False: 2},
        marks=pytest.mark.requires_ansible_v2(),
        id="8",
    ),
    # this is confusing, but how host slicing works on v2
    pytest.param(
        "*[0:]",
        {True: 6, False: 3},
        marks=pytest.mark.requires_ansible_v2(),
        id="9",
    ),
]

EXTRA_HOST_POSITIVE_PATTERNS = [
    # True means we are including the extra inventory, False not
    pytest.param(
        "extra-hosts",
        {True: 1, False: 0},
        marks=pytest.mark.requires_ansible_v2(),
        id="10",
    ),
    pytest.param(
        "extra-hosts_extra",
        {True: 1, False: 0},
        marks=pytest.mark.requires_ansible_v2(),
        id="11",
    ),
    pytest.param(
        "extra-host*",
        {True: 2, False: 0},
        marks=pytest.mark.requires_ansible_v2(),
        id="12",
    ),
]

NEGATIVE_HOST_PATTERNS = [
    # True means we are including the extra inventory, False not
    pytest.param("none", {True: 0, False: 0}, id="0"),
    pytest.param("all[8:]", {True: 0, False: 0}, id="1"),
]

POSITIVE_HOST_SLICES = [
    # True means we are including the extra inventory, False not
    pytest.param(slice(0, 0), {True: 2, False: 1}, id="0"),
    pytest.param(
        slice(0, 1),
        {True: 4, False: 2},
        marks=pytest.mark.requires_ansible_v2(),
        id="1",
    ),
    pytest.param(
        slice(0, 2),
        {True: 6, False: 3},
        marks=pytest.mark.requires_ansible_v2(),
        id="2",
    ),
    pytest.param(slice(0), {True: 2, False: 1}, id="3"),
    pytest.param(slice(1), {True: 2, False: 1}, id="4"),
    pytest.param(slice(2), {True: 2, False: 1}, id="5"),
    pytest.param(slice(3), {True: 2, False: 1}, id="6"),
]

NEGATIVE_HOST_SLICES = [
    slice(None),
    slice(-1),
]


def pytest_runtest_setup(item):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201, D103
    # Conditionally skip tests that are pinned to a specific ansible version
    if isinstance(item, pytest.Function):
        # conditionally xfail
        mark = item.get_closest_marker("ansible_v2_xfail")
        if mark:
            item.add_marker(
                pytest.xfail(
                    reason="expected failure on >= ansible-2.*",
                ),
            )


class PyTestOption:
    """Helper class that provides methods for creating and managing an inventory file."""

    def __init__(self, config, pytester) -> None:  # type: ignore[no-untyped-def]  # noqa: ANN001, D107
        self.config = config

        # Create inventory file
        self.inventory = pytester.makefile(
            ".ini",
            inventory="""
            [local]

            [local:children]
            reachable

            [reachable]
            localhost ansible_connection=local
            127.0.0.2 ansible_connection=local
            127.0.0.3 ansible_connection=local
            127.0.0.4 ansible_connection=local
            127.0.0.5 ansible_connection=local

            [unreachable]
            unreachable-host-1.example.com
            unreachable-host-2.example.com
            unreachable-host-3.example.com
        """,
        )

        # Create ansible.cfg file

    @property
    def args(self):  # type: ignore[no-untyped-def]  # noqa: ANN201, D102
        return ["--tb", "native"]


@pytest.fixture(autouse=True)
def _clear_global_context():  # type: ignore[no-untyped-def]  # noqa: ANN202
    # Reset the stored command line args
    # if context object does not exist because of old version of ansible, we don't need it
    if co is not None:
        co.GlobalCLIArgs._Singleton__instance = None


@pytest.fixture()
def option(request, pytester):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201
    """Returns an instance of PyTestOption to help tests pass parameters and
    use a common inventory file.
    """  # noqa: D205
    return PyTestOption(request.config, pytester)


@pytest.fixture()
def hosts():  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    def create_host_manager(include_extra_inventory=False):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN202, FBT002
        kwargs = {"inventory": ",".join(ALL_HOSTS), "connection": "local"}
        if include_extra_inventory:
            kwargs["extra_inventory"] = ",".join(ALL_EXTRA_HOSTS)
        return get_host_manager(**kwargs)

    return create_host_manager



"""Fixme."""

from __future__ import annotations


class ModuleResult(dict):  # type: ignore[type-arg]
    """Fixme."""

    def _check_key(self, key):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN202
        # if 'results' in self:
        #     for res in self.get('results', []):
        #         if isinstance(res, dict):
        return self.get(key, False)

    @property
    def is_ok(self):  # type: ignore[no-untyped-def]  # noqa: ANN201
        """Fixme."""
        return not (self.is_changed or self.is_unreachable or self.is_skipped or self.is_failed)

    @property
    def is_changed(self):  # type: ignore[no-untyped-def]  # noqa: ANN201
        """Fixme."""
        return self._check_key("changed")  # type: ignore[no-untyped-call]

    @property
    def is_unreachable(self):  # type: ignore[no-untyped-def]  # noqa: ANN201
        """Fixme."""
        return self._check_key("unreachable")  # type: ignore[no-untyped-call]

    @property
    def is_skipped(self):  # type: ignore[no-untyped-def]  # noqa: ANN201
        """Fixme."""
        return self._check_key("skipped")  # type: ignore[no-untyped-call]

    @property
    def is_failed(self):  # type: ignore[no-untyped-def]  # noqa: ANN201
        """Fixme."""
        return self._check_key("failed") or self.get("rc", 0) != 0  # type: ignore[no-untyped-call]

    @property
    def is_successful(self):  # type: ignore[no-untyped-def]  # noqa: ANN201
        """Fixme."""
        return not (self.is_failed or self.is_unreachable)


class AdHocResult:
    """Fixme."""

    def __init__(self, **kwargs) -> None:  # type: ignore[no-untyped-def]  # noqa: ANN003
        """Fixme."""
        required_kwargs = ("contacted",)
        for kwarg in required_kwargs:
            assert kwarg in kwargs, f"Missing required keyword argument '{kwarg}'"  # noqa: S101
            setattr(self, kwarg, kwargs.get(kwarg))

    def __getitem__(self, item):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN204
        """Return a ModuleResult instance matching the provided `item`."""
        if item in self.contacted:
            return ModuleResult(**self.contacted[item])
        raise KeyError(item)

    def __getattr__(self, attr):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN204
        """Return a ModuleResult instance matching the provided `attr`."""
        if attr in self.contacted:
            return ModuleResult(**self.contacted[attr])
        msg = f"type AdHocResult has no attribute '{attr}'"
        raise AttributeError(msg)

    def __len__(self) -> int:
        """Return the number of contacted hosts."""
        return len(self.contacted)

    def __contains__(self, item) -> bool:  # type: ignore[no-untyped-def]  # noqa: ANN001
        """Return whether the provided `item` was contacted."""
        return item in self.contacted

    def __iter__(self):  # type: ignore[no-untyped-def]  # noqa: ANN204
        """Return an iterator of the contacted inventory hosts."""
        return iter(self.contacted)

    def keys(self):  # type: ignore[no-untyped-def]  # noqa: ANN201
        """Return a list of contacted inventory hosts."""
        return self.contacted.keys()

    def items(self):  # type: ignore[no-untyped-def]  # noqa: ANN201
        """Return a list of tuples containing the inventory host key, and the ModuleResult instance."""  # noqa: E501
        for k in self.contacted:
            yield (k, getattr(self, k))

    def values(self):  # type: ignore[no-untyped-def]  # noqa: ANN201
        """Return a list of ModuleResult instances for each contacted inventory host."""
        return [getattr(self, k) for k in self.contacted]



"""Test the plugin."""

from __future__ import annotations

from unittest import mock
from unittest.mock import MagicMock

from pytest_ansible.plugin import PyTestAnsiblePlugin, pytest_generate_tests


class MockItem:
    """Mock class for item object."""

    def __init__(self, fixturenames, marker=None) -> None:  # type: ignore[no-untyped-def]  # noqa: ANN001, D107
        self.fixturenames = fixturenames
        self.marker = marker

    def get_closest_marker(self, marker_name):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201, ARG002, D102
        return self.marker


class MockConfig:
    """Mock class for config object.

    Attributes:
        options: A dictionary of options.
    """

    options = {}  # type: ignore[var-annotated]  # noqa: RUF012

    def setoption(self, option_name, value):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201, D102
        self.options[option_name] = value

    def getoption(self, option_name):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201, D102
        return self.options.get(option_name)

    def __init__(self) -> None:  # noqa: D107
        self.options = {
            "ansible_host_pattern": "localhost",
            "ansible_inventory": "/etc/ansible/hosts",
        }


class MockPluginManager:
    """Mock class for pluginmanager object."""

    def getplugin(self, name):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201, ARG002, D102
        return MagicMock()


class MockMetafunc:
    """Mock class for metafunc object."""

    def __init__(self, fixturenames) -> None:  # type: ignore[no-untyped-def]  # noqa: ANN001, D107
        self.fixturenames = fixturenames
        self.config = MockConfig()
        self.parametrize = MagicMock()


def test_pytest_generate_tests_with_ansible_host():  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    metafunc = MagicMock()
    metafunc.fixturenames = ["ansible_host"]
    metafunc.config = MagicMock()

    # Mock config values
    metafunc.config.getoption.side_effect = {
        "ansible_host_pattern": "localhost",
        "ansible_inventory": "/etc/ansible/hosts",
    }.get

    plugin = PyTestAnsiblePlugin(metafunc.config)

    # Mock Ansible host initialization
    host = MagicMock()
    host.name = "localhost"
    plugin.initialize = MagicMock(return_value={"localhost": host})  # type: ignore[method-assign]

    pytest_generate_tests(metafunc)  # type: ignore[no-untyped-call]

    assert metafunc.parametrize.call_count == 1


def test_pytest_generate_tests_with_ansible_group():  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    metafunc = MagicMock()
    metafunc.fixturenames = ["ansible_group"]
    config = MagicMock()
    config.pluginmanager = MagicMock()
    metafunc.config = config

    plugin = PyTestAnsiblePlugin(metafunc.config)

    # Mock Ansible host initialization
    host1 = MagicMock()
    host1.name = "host1"
    group1 = MagicMock()
    group1.name = "group1"
    host1.groups = [group1]

    host2 = MagicMock()
    host2.name = "host2"
    group2 = MagicMock()
    group2.name = "group2"
    host2.groups = [group2]

    plugin.initialize = MagicMock(return_value={"host1": host1, "host2": host2})  # type: ignore[method-assign]

    pytest_generate_tests(metafunc)  # type: ignore[no-untyped-call]

    assert metafunc.parametrize.call_count == 2  # noqa: PLR2004  # Called twice for ansible_group


def test_pytest_collection_modifyitems_with_marker():  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    # Mock configuration with ansible_ marker
    mock_config = MockConfig()
    mock_config.setoption("ansible_host_pattern", "some_pattern")  # type: ignore[no-untyped-call]
    mock_config.setoption("ansible_inventory", "some_inventory")  # type: ignore[no-untyped-call]

    plugin = PyTestAnsiblePlugin(mock_config)
    items = [
        MockItem(
            fixturenames=["ansible_fixture"],
            marker=mock.Mock(name="ansible_marker"),
        ),
    ]

    # With the marker, ensure that assert_required_ansible_parameters is not called
    with mock.patch.object(plugin, "assert_required_ansible_parameters"):
        plugin.pytest_collection_modifyitems(None, mock_config, items)  # type: ignore[no-untyped-call]


def test_pytest_collection_modifyitems_without_marker():  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    # Mock configuration without ansible_ marker
    mock_config = MockConfig()
    mock_config.setoption("ansible_host_pattern", "some_pattern")  # type: ignore[no-untyped-call]
    mock_config.setoption("ansible_inventory", "some_inventory")  # type: ignore[no-untyped-call]

    plugin = PyTestAnsiblePlugin(mock_config)
    items = [MockItem(fixturenames=["ansible_adhoc"])]

    # Without the marker, ensure that assert_required_ansible_parameters is called
    with mock.patch.object(plugin, "assert_required_ansible_parameters") as mock_assert:
        plugin.pytest_collection_modifyitems(None, mock_config, items)  # type: ignore[no-untyped-call]
        mock_assert.assert_called_once()


def test_pytest_collection_modifyitems_no_fixtures():  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    # Mock configuration without ansible_ marker
    mock_config = MockConfig()
    mock_config.setoption("ansible_host_pattern", "some_pattern")  # type: ignore[no-untyped-call]
    mock_config.setoption("ansible_inventory", "some_inventory")  # type: ignore[no-untyped-call]

    plugin = PyTestAnsiblePlugin(mock_config)
    items = [MockItem(fixturenames=[])]

    # With no fixtures, ensure that assert_required_ansible_parameters is not called
    with mock.patch.object(plugin, "assert_required_ansible_parameters") as mock_assert:
        plugin.pytest_collection_modifyitems(None, mock_config, items)  # type: ignore[no-untyped-call]
        mock_assert.assert_not_called()



"""Ad-hoc tests for pytest-ansible."""

from __future__ import annotations

import pytest


# pylint: disable=unused-import
try:
    from _pytest.main import (  # type: ignore[attr-defined]
        EXIT_INTERRUPTED,
        EXIT_NOTESTSCOLLECTED,
        EXIT_OK,
        EXIT_TESTSFAILED,
        EXIT_USAGEERROR,
    )
except ImportError:
    from _pytest.main import ExitCode  # type: ignore[attr-defined]

    EXIT_OK = ExitCode.OK
    EXIT_TESTSFAILED = ExitCode.TESTS_FAILED
    EXIT_USAGEERROR = ExitCode.USAGE_ERROR
    EXIT_INTERRUPTED = ExitCode.INTERRUPTED
    EXIT_NOTESTSCOLLECTED = ExitCode.NO_TESTS_COLLECTED


@pytest.mark.old()
def test_contacted_with_params(pytester, option):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201
    """FIXME."""
    src = """
        import pytest
        def test_func(ansible_module):
            contacted = ansible_module.ping()

            # assert contacted hosts ...
            assert contacted
            assert len(contacted) == len(ansible_module)
            for result in contacted.values():
                assert result.is_successful
                assert result['ping'] == 'pong'

    """
    pytester.makepyfile(src)
    result = pytester.runpytest_subprocess(
        *[
            *option.args,
            "--ansible-inventory",
            str(option.inventory),
            "--ansible-host-pattern",
            "local",
        ],
    )
    assert result.ret == EXIT_OK
    assert result.parseoutcomes()["passed"] == 1


@pytest.mark.old()
def test_contacted_with_params_and_inventory_marker(pytester, option):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201
    """FIXME."""
    src = f"""
        import pytest
        @pytest.mark.ansible(inventory='{option.inventory}')
        def test_func(ansible_module):
            contacted = ansible_module.ping()

            # assert contacted hosts ...
            assert contacted
            assert len(contacted) == len(ansible_module)
            for result in contacted.values():
                assert result.is_successful
                assert result['ping'] == 'pong'

        """

    pytester.makepyfile(src)
    result = pytester.runpytest_subprocess(
        *[*option.args, "--ansible-host-pattern", "local"],
    )
    assert result.ret == EXIT_OK
    assert result.parseoutcomes()["passed"] == 1


@pytest.mark.old()
def test_contacted_with_params_and_host_pattern_marker(pytester, option):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201
    """FIXME."""
    src = """
        import pytest
        @pytest.mark.ansible(host_pattern='local')
        def test_func(ansible_module):
            contacted = ansible_module.ping()

            # assert contacted hosts ...
            assert contacted
            assert len(contacted) == len(ansible_module)
            for result in contacted.values():
                assert result.is_successful
                assert result['ping'] == 'pong'

    """
    pytester.makepyfile(src)
    result = pytester.runpytest_subprocess(
        *[
            *option.args,
            "--ansible-inventory",
            str(option.inventory),
            "--ansible-host-pattern",
            "unreachable",
        ],
    )
    assert result.ret == EXIT_OK
    assert result.parseoutcomes()["passed"] == 1


@pytest.mark.old()
def test_contacted_with_params_and_inventory_host_pattern_marker(pytester, option):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201
    """FIXME."""
    src = f"""
        import pytest
        @pytest.mark.ansible(inventory='{option.inventory}', host_pattern='local')
        def test_func(ansible_module):
            contacted = ansible_module.ping()

            # assert contacted hosts ...
            assert contacted
            assert len(contacted) == len(ansible_module)
            for result in contacted.values():
                assert result.is_successful
                assert result['ping'] == 'pong'
        """

    pytester.makepyfile(src)
    result = pytester.runpytest_subprocess(
        *[
            *option.args,
            "--ansible-inventory",
            str(option.inventory),
            "--ansible-host-pattern",
            "unreachable",
        ],
    )
    assert result.ret == EXIT_OK
    assert result.parseoutcomes()["passed"] == 1


@pytest.mark.old()
def test_become(pytester, option):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201
    """Test --ansible-become* parameters.  This test doesn't actually 'sudo',
    but verifies that 'sudo' was attempted by asserting
    '--ansible-become-user' fails as expected.
    """  # noqa: D205
    src = f"""
        import pytest
        import ansible
        import re
        import os
        from packaging.version import parse as parse_version

        @pytest.mark.ansible(inventory='{option.inventory}', host_pattern='localhost')
        def test_func(ansible_module):
            contacted = ansible_module.ping()
            # assert contacted hosts ...
            assert contacted
            assert len(contacted) == len(ansible_module)
            for result in contacted.values():
                # Assert test failed as expected
                if parse_version(ansible.__version__) < parse_version('2.4.0'):
                    assert 'failed' in result, "Missing expected field in JSON response: failed"
                    assert result['failed'], "Test did not fail as expected"

                # Assert expected failure message
                if parse_version(ansible.__version__) >= parse_version('2.0.0'):
                    assert 'msg' in result, "Missing expected field in JSON response: msg"
                    assert result['msg'].startswith('Failed to set permissions on the temporary files Ansible needs ' \
                        'to create when becoming an unprivileged user')
                else:
                    assert 'msg' in result, "Missing expected field in JSON response: msg"
                    assert 'sudo: unknown user: unknown_user' in result['msg']
        """  # noqa: E501

    pytester.makepyfile(src)
    result = pytester.runpytest_subprocess(
        *option.args  # noqa: RUF005
        + [
            "--ansible-inventory",
            str(option.inventory),
            "--ansible-host-pattern",
            "localhost",  # run against a single host
            "--ansible-become",  # Enable become support
            "--ansible-become-user",
            "unknown_user",  # Connect as unknown_user
        ],
    )
    assert result.ret == EXIT_OK
    assert result.parseoutcomes()["passed"] == 1


@pytest.mark.old()
def test_dark_with_params(pytester, option):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201
    """FIXME."""
    src = """
        import pytest
        from pytest_ansible.errors import (AnsibleConnectionFailure, AnsibleNoHostsMatch)
        def test_func(ansible_module):
            exc_info = pytest.raises(AnsibleConnectionFailure, ansible_module.ping)

            # assert no contacted hosts ...
            assert not exc_info.value.contacted, "%d hosts were contacted, expected %d" \
                % (len(exc_info.value.contacted), 0)

            # assert dark hosts ...
            assert exc_info.value.dark
    """
    pytester.makepyfile(src)
    result = pytester.runpytest_subprocess(
        *[
            *option.args,
            "--ansible-inventory",
            str(option.inventory),
            "--ansible-host-pattern",
            "unreachable",
        ],
    )
    assert result.ret == EXIT_OK
    assert result.parseoutcomes()["passed"] == 1


@pytest.mark.old()
def test_dark_with_params_and_inventory_marker(pytester, option):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201
    """FIXME."""
    src = f"""
        import pytest
        from pytest_ansible.errors import (AnsibleConnectionFailure, AnsibleNoHostsMatch)
        @pytest.mark.ansible(inventory='{option.inventory}')
        def test_func(ansible_module):
            exc_info = pytest.raises(AnsibleConnectionFailure, ansible_module.ping)

            # assert no contacted hosts ...
            assert not exc_info.value.contacted, "%d hosts were contacted, expected %d" \
                % (len(exc_info.value.contacted), 0)

            # assert dark hosts ...
            assert exc_info.value.dark
        """

    pytester.makepyfile(src)
    result = pytester.runpytest_subprocess(
        *[*option.args, "--ansible-host-pattern", "unreachable"],
    )
    assert result.ret == EXIT_OK
    assert result.parseoutcomes()["passed"] == 1


@pytest.mark.old()
def test_dark_with_params_and_host_pattern_marker(pytester, option):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201
    """FIXME."""
    src = """
        import pytest
        import ansible
        from pytest_ansible.errors import (AnsibleConnectionFailure, AnsibleNoHostsMatch)
        @pytest.mark.ansible(host_pattern='unreachable')
        def test_func(ansible_module):
            exc_info = pytest.raises(AnsibleConnectionFailure, ansible_module.ping)

            # assert no contacted hosts ...
            assert not exc_info.value.contacted, "%d hosts were contacted, expected %d" \
                % (len(exc_info.value.contacted), 0)

            # assert dark hosts ...
            assert exc_info.value.dark
    """
    pytester.makepyfile(src)
    result = pytester.runpytest_subprocess(
        *[
            *option.args,
            "--ansible-inventory",
            str(option.inventory),
            "--ansible-host-pattern",
            "local",
        ],
    )
    assert result.ret == EXIT_OK
    assert result.parseoutcomes()["passed"] == 1


@pytest.mark.old()
def test_dark_with_debug_enabled(pytester, option):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201
    """Verify that when verbosity is enabled, additional output is provided upon host failure."""
    src = """
        import pytest
        from pytest_ansible.errors import AnsibleConnectionFailure
        def test_func(ansible_module):
            ansible_module.ping()
    """
    pytester.makepyfile(src)
    result = pytester.runpytest_subprocess(
        *[
            *option.args,
            "--ansible-inventory",
            str(option.inventory),
            "--ansible-host-pattern",
            "unreachable",
            "-v",
        ],
    )
    assert result.ret == EXIT_TESTSFAILED
    assert result.parseoutcomes()["failed"] == 1



"""Fixme."""

from __future__ import annotations

from ansible.inventory.manager import InventoryManager
from ansible.parsing.dataloader import DataLoader
from ansible.vars.manager import VariableManager

from pytest_ansible.host_manager.base import BaseHostManager
from pytest_ansible.module_dispatcher.v213 import ModuleDispatcherV213


class HostManagerV213(BaseHostManager):
    """Fixme."""

    def __init__(self, *args, **kwargs) -> None:  # type: ignore[no-untyped-def]  # noqa: ANN002, ANN003
        """Fixme."""
        super().__init__(*args, **kwargs)
        self._dispatcher = ModuleDispatcherV213

    def initialize_inventory(self):  # type: ignore[no-untyped-def]  # noqa: ANN201
        """Fixme."""
        self.options["loader"] = DataLoader()
        self.options["inventory_manager"] = InventoryManager(
            loader=self.options["loader"],
            sources=self.options["inventory"],
        )
        self.options["variable_manager"] = VariableManager(
            loader=self.options["loader"],
            inventory=self.options["inventory_manager"],
        )
        if self.options.get("extra_inventory", None):
            self.options["extra_loader"] = DataLoader()
            self.options["extra_inventory_manager"] = InventoryManager(
                loader=self.options["extra_loader"],
                sources=self.options["extra_inventory"],
            )
            self.options["extra_variable_manager"] = VariableManager(
                loader=self.options["extra_loader"],
                inventory=self.options["extra_inventory_manager"],
            )



"""Fixme."""

from __future__ import annotations

import sys
import warnings

import ansible.constants
import ansible.errors
import ansible.utils

from ansible.cli.adhoc import AdHocCLI
from ansible.constants import COLLECTIONS_PATHS  # pylint: disable=no-name-in-module
from ansible.executor.task_queue_manager import TaskQueueManager
from ansible.playbook.play import Play
from ansible.plugins.callback import CallbackBase
from ansible.plugins.loader import module_loader

from pytest_ansible.errors import AnsibleConnectionFailure
from pytest_ansible.has_version import has_ansible_v213
from pytest_ansible.module_dispatcher import BaseModuleDispatcher
from pytest_ansible.results import AdHocResult


HAS_CUSTOM_LOADER_SUPPORT = True

try:
    # init_plugin_loader was introduced in Ansible-core change here, v2.15
    # https://github.com/ansible/ansible/pull/78915
    # Whenever a new vXYZ.py dispatcher module is introduced, make this static import
    from ansible.plugins.loader import init_plugin_loader
except ImportError:
    HAS_CUSTOM_LOADER_SUPPORT = False


class ResultAccumulator(CallbackBase):  # type: ignore[misc]
    """Fixme."""

    def __init__(self, *args, **kwargs) -> None:  # type: ignore[no-untyped-def]  # noqa: ANN002, ANN003
        """Initialize object."""
        super().__init__(*args, **kwargs)
        self.contacted = {}  # type: ignore[var-annotated]
        self.unreachable = {}  # type: ignore[var-annotated]

    def v2_runner_on_failed(self, result, *args, **kwargs):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN002, ANN003, ANN201, ARG002
        """Fixme."""
        result2 = {"failed": True}
        result2.update(result._result)  # noqa: SLF001
        self.contacted[result._host.get_name()] = result2  # noqa: SLF001

    def v2_runner_on_ok(self, result):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201
        """Fixme."""
        self.contacted[result._host.get_name()] = result._result  # noqa: SLF001

    def v2_runner_on_unreachable(self, result):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201
        """Fixme."""
        self.unreachable[result._host.get_name()] = result._result  # noqa: SLF001

    @property
    def results(self):  # type: ignore[no-untyped-def]  # noqa: ANN201
        """Fixme."""
        return {"contacted": self.contacted, "unreachable": self.unreachable}


class ModuleDispatcherV213(BaseModuleDispatcher):
    """A plugin runner for Ansible 2.13 and newer.

    Attributes:
        required_kwargs: A tuple of required keyword arguments.
    """

    required_kwargs = (
        "inventory",
        "inventory_manager",
        "variable_manager",
        "host_pattern",
        "loader",
    )

    def __init__(self, **kwargs) -> None:  # type: ignore[no-untyped-def]  # noqa: ANN003
        """Fixme."""
        super().__init__(**kwargs)
        if not has_ansible_v213:
            msg = "Only supported with ansible-2.13 and newer"
            raise ImportError(msg)

    def has_module(self: ModuleDispatcherV213, name: str) -> str:
        """Determine if a module exists and return the full name or "".

        Attributes:
            name: The user provided name of the module to check.

        Returns:
            The full name of the module if it exists or "" if it does not.
        """
        if "module_path" in self.options:
            paths = self.options["module_path"]
            if isinstance(paths, list | tuple | set):
                for path in paths:
                    module_loader.add_directory(path)
            else:
                module_loader.add_directory(paths)
        try:
            found = module_loader.find_plugin_with_context(name)
        except ModuleNotFoundError:
            return ""
        if not getattr(found, "resolved", False):
            return ""
        return str(found.resolved_fqcn)

    def _run(self, *module_args, **complex_args):  # type: ignore[no-untyped-def]  # noqa: ANN002, ANN003, ANN202, C901, PLR0912, PLR0915
        """Execute an ansible adhoc command returning the result in a AdhocResult object."""
        # Assemble module argument string
        if module_args:
            complex_args.update({"_raw_params": " ".join(module_args)})

        # Assert hosts matching the provided pattern exist
        hosts = self.options["inventory_manager"].list_hosts()
        if self.options.get("extra_inventory_manager", None):
            extra_hosts = self.options["extra_inventory_manager"].list_hosts()
        else:
            extra_hosts = []
        no_hosts = False
        if len(hosts + extra_hosts) == 0:
            no_hosts = True
            warnings.warn("provided hosts list is empty, only localhost is available")  # noqa: B028

        self.options["inventory_manager"].subset(self.options.get("subset"))
        hosts = self.options["inventory_manager"].list_hosts(
            self.options["host_pattern"],
        )
        if self.options.get("extra_inventory_manager", None):
            self.options["extra_inventory_manager"].subset(self.options.get("subset"))
            extra_hosts = self.options["extra_inventory_manager"].list_hosts()
        else:
            extra_hosts = []
        if len(hosts + extra_hosts) == 0 and not no_hosts:
            msg = "Specified hosts and/or --limit does not match any hosts."
            raise ansible.errors.AnsibleError(
                msg,
            )

        # Pass along cli options
        args = ["pytest-ansible"]
        verbosity = None
        for verbosity_syntax in ("-v", "-vv", "-vvv", "-vvvv", "-vvvvv"):
            if verbosity_syntax in sys.argv:
                verbosity = verbosity_syntax
                break
        if verbosity is not None:
            args.append(verbosity_syntax)
        args.extend([self.options["host_pattern"]])
        for argument in (
            "connection",
            "user",
            "become",
            "become_method",
            "become_user",
            "module_path",
        ):
            arg_value = self.options.get(argument)
            argument = argument.replace("_", "-")  # noqa: PLW2901

            if arg_value in (None, False):
                continue

            if arg_value is True:
                args.append(f"--{argument}")
            else:
                args.append(f"--{argument}={arg_value}")

        # Use Ansible's own adhoc cli to parse the fake command line we created and then save it
        # into Ansible's global context
        adhoc = AdHocCLI(args)
        adhoc.parse()

        # And now we'll never speak of this again
        del adhoc

        # Initialize callbacks to capture module JSON responses
        callback = ResultAccumulator()

        kwargs = {
            "inventory": self.options["inventory_manager"],
            "variable_manager": self.options["variable_manager"],
            "loader": self.options["loader"],
            "stdout_callback": callback,
            "passwords": {"conn_pass": None, "become_pass": None},
        }

        kwargs_extra = {}
        # If we have an extra inventory, do the same that we did for the inventory
        if self.options.get("extra_inventory_manager", None):
            callback_extra = ResultAccumulator()

            kwargs_extra = {
                "inventory": self.options["extra_inventory_manager"],
                "variable_manager": self.options["extra_variable_manager"],
                "loader": self.options["extra_loader"],
                "stdout_callback": callback_extra,
                "passwords": {"conn_pass": None, "become_pass": None},
            }

        # create a pseudo-play to execute the specified module via a single task
        play_ds = {
            "name": "pytest-ansible",
            "hosts": self.options["host_pattern"],
            "become": self.options.get("become"),
            "become_user": self.options.get("become_user"),
            "gather_facts": "no",
            "tasks": [
                {
                    "action": {
                        "module": self.options["module_name"],
                        "args": complex_args,
                    },
                },
            ],
        }

        play = Play().load(
            play_ds,
            variable_manager=self.options["variable_manager"],
            loader=self.options["loader"],
        )
        play_extra = None
        if self.options.get("extra_inventory_manager", None):
            play_extra = Play().load(
                play_ds,
                variable_manager=self.options["extra_variable_manager"],
                loader=self.options["extra_loader"],
            )

        if HAS_CUSTOM_LOADER_SUPPORT:
            # Load the collection finder, unsupported, may change in future
            init_plugin_loader(COLLECTIONS_PATHS)

        # now create a task queue manager to execute the play
        tqm = None
        try:
            tqm = TaskQueueManager(**kwargs)
            tqm.run(play)
        finally:
            if tqm:
                tqm.cleanup()

        if self.options.get("extra_inventory_manager", None):
            tqm_extra = None
            try:
                tqm_extra = TaskQueueManager(**kwargs_extra)
                tqm_extra.run(play_extra)
            finally:
                if tqm_extra:
                    tqm_extra.cleanup()

        # Raise exception if host(s) unreachable
        if callback.unreachable:
            msg = "Host unreachable in the inventory"
            raise AnsibleConnectionFailure(
                msg,
                dark=callback.unreachable,
                contacted=callback.contacted,
            )
        if self.options.get("extra_inventory_manager", None) and callback_extra.unreachable:
            raise AnsibleConnectionFailure(  # noqa: TRY003
                "Host unreachable in the extra inventory",  # noqa: EM101
                dark=callback_extra.unreachable,
                contacted=callback_extra.contacted,
            )

        # Success!
        return AdHocResult(
            contacted=(
                {**callback.contacted, **callback_extra.contacted}
                if self.options.get("extra_inventory_manager", None)
                else callback.contacted
            ),
        )



"""Test the module dispatcher."""

from __future__ import annotations

import pytest

from .conftest import NEGATIVE_HOST_PATTERNS, POSITIVE_HOST_PATTERNS


def test_type_error() -> None:
    """Verify that BaseModuleDispatcher cannot be instantiated."""
    from pytest_ansible.module_dispatcher import BaseModuleDispatcher

    with pytest.raises(TypeError, match="^Can't instantiate.*$"):
        BaseModuleDispatcher(inventory="localhost,")  # type: ignore[abstract] #pylint: disable=abstract-class-instantiated


@pytest.mark.requires_ansible_v2()
def test_importerror_requires_v1():  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    with pytest.raises(ImportError):
        # pylint: disable=unused-import
        import pytest_ansible.module_dispatcher.v1  # type: ignore[import-not-found] # noqa: F401 # pylint: disable=import-error, no-name-in-module


@pytest.mark.parametrize(
    ("host_pattern", "num_hosts"),
    POSITIVE_HOST_PATTERNS,
)
@pytest.mark.parametrize(
    "include_extra_inventory",
    (True, False),
)
def test_dispatcher_len(host_pattern, num_hosts, hosts, include_extra_inventory):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201, D103
    hosts = hosts(include_extra_inventory=include_extra_inventory)
    assert len(getattr(hosts, host_pattern)) == num_hosts[include_extra_inventory]


@pytest.mark.parametrize(
    ("host_pattern", "num_hosts"),
    POSITIVE_HOST_PATTERNS,
)
@pytest.mark.parametrize(
    "include_extra_inventory",
    (True, False),
)
def test_dispatcher_contains(host_pattern, num_hosts, hosts, include_extra_inventory):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201, ARG001, D103
    hosts = hosts(include_extra_inventory=include_extra_inventory)
    assert host_pattern in hosts["all"]


@pytest.mark.parametrize(("host_pattern", "num_hosts"), NEGATIVE_HOST_PATTERNS)
@pytest.mark.parametrize(
    "include_extra_inventory",
    (True, False),
)
def test_dispatcher_not_contains(  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    host_pattern,  # noqa: ANN001
    num_hosts,  # noqa: ANN001, ARG001
    hosts,  # noqa: ANN001
    include_extra_inventory,  # noqa: ANN001
):
    hosts = hosts(include_extra_inventory=include_extra_inventory)
    assert host_pattern not in hosts["all"]


def test_ansible_module_error(hosts):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201
    """Verify that AnsibleModuleError is raised when no such module exists."""
    from pytest_ansible.errors import AnsibleModuleError

    with pytest.raises(AnsibleModuleError) as exc_info:
        hosts().all.a_module_that_most_certainly_does_not_exist()
    assert (
        str(exc_info.value)
        == f"The module {'a_module_that_most_certainly_does_not_exist'} was not found in configured module paths."  # noqa: E501
    )



"""Fixme."""

from __future__ import annotations

import ansible

from packaging.version import parse as parse_version


has_ansible_v2 = parse_version(ansible.__version__) >= parse_version("2.0.0")
has_ansible_v24 = parse_version(ansible.__version__) >= parse_version("2.4.0")
has_ansible_v28 = parse_version(ansible.__version__) >= parse_version(
    "2.8.0.dev0",
) or parse_version(ansible.__version__) >= parse_version("2.8.0")
has_ansible_v29 = parse_version(ansible.__version__) >= parse_version("2.9.0")
has_ansible_v212 = parse_version(ansible.__version__) >= parse_version("2.12.0")
has_ansible_v213 = parse_version(ansible.__version__) >= parse_version("2.13.0")



"""Defines pytest-ansible exception classes."""

from __future__ import annotations

import ansible.errors


class AnsibleNoHostsMatch(ansible.errors.AnsibleError):  # type: ignore[misc]
    """Sub-class AnsibleError when no hosts match."""


class AnsibleConnectionFailure(ansible.errors.AnsibleError):  # type: ignore[misc]
    """Sub-class AnsibleError when connection failures occur."""

    def __init__(self, msg, dark=None, contacted=None) -> None:  # type: ignore[no-untyped-def]  # noqa: ANN001
        """Initialize connection error class."""
        super().__init__(msg)
        self.contacted = contacted
        self.dark = dark


class AnsibleModuleError(ansible.errors.AnsibleError):  # type: ignore[misc]
    """Sub-class AnsibleError when module failures occur."""



"""PyTest fixtures."""

from __future__ import annotations

import pytest


@pytest.fixture(name="ansible_adhoc")
def fixture_ansible_adhoc(request):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201
    """Return an inventory initialization method."""
    plugin = request.config.pluginmanager.getplugin("ansible")

    def init_host_mgr(**kwargs):  # type: ignore[no-untyped-def]  # noqa: ANN003, ANN202
        return plugin.initialize(request.config, request, **kwargs)

    return init_host_mgr


@pytest.fixture(name="ansible_module")
def fixture_ansible_module(ansible_adhoc):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201
    """Return a subclass of BaseModuleDispatcher."""
    host_mgr = ansible_adhoc()
    return getattr(host_mgr, host_mgr.options["host_pattern"])


@pytest.fixture()
def ansible_facts(ansible_module):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201
    """Return ansible_facts dictionary."""
    return ansible_module.setup()


@pytest.fixture()
def localhost(request):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201
    """Return a host manager representing localhost."""
    # NOTE: Do not use ansible_adhoc as a dependent fixture since that will assert specific command-line parameters have  # noqa: E501
    # been supplied.  In the case of localhost, the parameters are provided as kwargs below.
    plugin = request.config.pluginmanager.getplugin("ansible")
    return plugin.initialize(
        request.config,
        request,
        inventory="localhost,",
        connection="local",
        host_pattern="localhost",
    ).localhost



"""Test CLI parameters."""

from __future__ import annotations

import ansible
import pytest


# pylint: disable=unused-import
try:
    from _pytest.main import (  # type: ignore[attr-defined]
        EXIT_INTERRUPTED,
        EXIT_NOTESTSCOLLECTED,
        EXIT_OK,
        EXIT_TESTSFAILED,
        EXIT_USAGEERROR,
    )
except ImportError:
    from _pytest.main import ExitCode  # type: ignore[attr-defined]

    EXIT_OK = ExitCode.OK
    EXIT_TESTSFAILED = ExitCode.TESTS_FAILED
    EXIT_USAGEERROR = ExitCode.USAGE_ERROR
    EXIT_INTERRUPTED = ExitCode.INTERRUPTED
    EXIT_NOTESTSCOLLECTED = ExitCode.NO_TESTS_COLLECTED


def test_plugin_help(pytester):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201
    """Verifies expected output from of pytest --help."""
    result = pytester.runpytest("--help")
    result.stdout.fnmatch_lines(
        [
            # Check for the github args section header
            "pytest-ansible:",
            # Check for the specific args
            "  --inventory=ANSIBLE_INVENTORY, --ansible-inventory=ANSIBLE_INVENTORY",
            "  --host-pattern=ANSIBLE_HOST_PATTERN, --ansible-host-pattern=ANSIBLE_HOST_PATTERN",
            "  --connection=ANSIBLE_CONNECTION, --ansible-connection=ANSIBLE_CONNECTION",
            "  --user=ANSIBLE_USER, --ansible-user=ANSIBLE_USER",
            "  --check, --ansible-check",
            "  --module-path=ANSIBLE_MODULE_PATH, --ansible-module-path=ANSIBLE_MODULE_PATH",
            "  --become, --ansible-become",
            "  --become-method=ANSIBLE_BECOME_METHOD, --ansible-become-method=ANSIBLE_BECOME_METHOD",  # noqa: E501
            "  --become-user=ANSIBLE_BECOME_USER, --ansible-become-user=ANSIBLE_BECOME_USER",
            "  --ask-become-pass=ANSIBLE_ASK_BECOME_PASS, --ansible-ask-become-pass=ANSIBLE_ASK_BECOME_PASS",  # noqa: E501
            # Check for the marker in --help
            "  ansible (args)*Ansible integration",
        ],
    )


def test_plugin_markers(pytester):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201
    """Verifies expected output from of pytest --markers."""
    result = pytester.runpytest("--markers")
    result.stdout.fnmatch_lines(
        [
            "@pytest.mark.ansible(*args): Ansible integration",
        ],
    )


def test_report_header(pytester, option):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201
    """Verify the expected ansible version in the pytest report header."""
    result = pytester.runpytest(*option.args)
    assert result.ret == EXIT_NOTESTSCOLLECTED
    result.stdout.fnmatch_lines([f"ansible: {ansible.__version__}"])


def test_params_not_required_when_not_using_fixture(pytester, option):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201
    """Verify the ansible parameters are not required if the fixture is not used."""
    src = """
        import pytest
        def test_func():
            assert True
    """
    pytester.makepyfile(src)
    result = pytester.runpytest(*option.args)
    assert result.ret == EXIT_OK


@pytest.mark.parametrize(
    "fixture_name",
    (
        "ansible_adhoc",
        "ansible_module",
        "ansible_facts",
    ),
)
def test_params_required_when_using_fixture(pytester, option, fixture_name):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201
    """Verify the ansible parameters are required if the fixture is used."""
    src = f"""
        import pytest
        def test_func({fixture_name}):
            {fixture_name}
        """

    pytester.makepyfile(src)
    result = pytester.runpytest(*option.args)
    assert result.ret == EXIT_USAGEERROR
    result.stderr.fnmatch_lines(
        [
            "ERROR: Missing required parameter --ansible-host-pattern/--host-pattern",
        ],
    )


@pytest.mark.parametrize(
    "required_value_parameter",
    (
        "--ansible-inventory",
        "--inventory",
        "--ansible-host-pattern",
        "--host-pattern",
        "--ansible-connection",
        "--connection",
        "--ansible-user",
        "--user",
        "--ansible-become-method",
        "--become-method",
        "--ansible-become-user",
        "--become-user",
        "--ansible-module-path",
        "--module-path",
    ),
)
def test_param_requires_value(pytester, required_value_parameter):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201
    """Verifies failure when not providing a value to a parameter that requires a value."""
    result = pytester.runpytest(*[required_value_parameter])
    assert result.ret == EXIT_USAGEERROR
    result.stderr.fnmatch_lines(
        [f"*: error: argument *{required_value_parameter}*: expected one argument"],
    )


def test_params_required_with_inventory_without_host_pattern(pytester, option):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201
    """Verify that a host pattern is required when an inventory is supplied."""
    src = """
        import pytest
        def test_func(ansible_module):
            assert True
    """
    pytester.makepyfile(src)
    result = pytester.runpytest(*[*option.args, "--ansible-inventory", "local,"])
    assert result.ret == EXIT_USAGEERROR
    result.stderr.fnmatch_lines(
        [
            "ERROR: Missing required parameter --ansible-host-pattern/--host-pattern",
        ],
    )


@pytest.mark.requires_ansible_v2()
def test_params_required_without_inventory_with_host_pattern_v2(pytester, option):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201, D103
    src = """
        import pytest
        def test_func(ansible_module):
            assert True
    """
    pytester.makepyfile(src)
    result = pytester.runpytest(*[*option.args, "--ansible-host-pattern", "all"])
    assert result.ret == EXIT_OK


def test_param_override_with_marker(pytester, option):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201, D103
    src = """
        import pytest
        @pytest.mark.ansible(inventory='local,', connection='local', host_pattern='all')
        def test_func(ansible_module):
            ansible_module.ping()
    """
    pytester.makepyfile(src)
    result = pytester.runpytest(
        *[
            *option.args,
            "--tb",
            "native",
            "--ansible-inventory",
            "garbage,",
            "--ansible-host-pattern",
            "garbage",
            "--ansible-connection",
            "garbage",
        ],
    )
    assert result.ret == EXIT_OK

    # Mock assert the correct variables are set



"""Test the fixtures provided by the plugin."""

from __future__ import annotations


try:
    from _pytest.main import EXIT_OK  # type: ignore  # noqa: PGH003
except ImportError:
    from _pytest.main import ExitCode  # type: ignore[attr-defined]

    EXIT_OK = ExitCode.OK


def test_ansible_adhoc(pytester, option):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201, D103
    src = """
        import pytest
        import types
        from pytest_ansible.host_manager.base import BaseHostManager
        def test_func(ansible_adhoc):
            assert isinstance(ansible_adhoc, types.FunctionType)
            assert isinstance(ansible_adhoc(), BaseHostManager)
    """
    pytester.makepyfile(src)
    result = pytester.runpytest(
        *[
            *option.args,
            "--ansible-inventory",
            str(option.inventory),
            "--ansible-host-pattern",
            "local",
        ],
    )
    assert result.ret == EXIT_OK
    assert result.parseoutcomes()["passed"] == 1


def test_ansible_module(pytester, option):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201, D103
    src = """
        import pytest
        from pytest_ansible.module_dispatcher import BaseModuleDispatcher
        def test_func(ansible_module):
            assert isinstance(ansible_module, BaseModuleDispatcher)
    """
    pytester.makepyfile(src)
    result = pytester.runpytest(
        *[
            *option.args,
            "--ansible-inventory",
            str(option.inventory),
            "--ansible-host-pattern",
            "local",
        ],
    )
    assert result.ret == EXIT_OK
    assert result.parseoutcomes()["passed"] == 1


def test_ansible_facts(pytester, option):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201, D103
    src = """
        import pytest
        from pytest_ansible.results import AdHocResult
        def test_func(ansible_facts):
            assert isinstance(ansible_facts, AdHocResult)
    """
    pytester.makepyfile(src)
    result = pytester.runpytest(
        *[
            *option.args,
            "--ansible-inventory",
            str(option.inventory),
            "--ansible-host-pattern",
            "local",
        ],
    )
    assert result.ret == EXIT_OK
    assert result.parseoutcomes()["passed"] == 1


def test_localhost(pytester, option):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201, D103
    src = """
        import pytest
        from pytest_ansible.module_dispatcher import BaseModuleDispatcher
        def test_func(localhost):
            assert isinstance(localhost, BaseModuleDispatcher)
    """
    pytester.makepyfile(src)
    result = pytester.runpytest(*option.args)
    assert result.ret == EXIT_OK
    assert result.parseoutcomes()["passed"] == 1
